@@ -1,15 +1,15 @@
-mod wrappers;
 mod commands;
 mod key;
+mod logging;
+mod wrappers;
 
-use std::ffi::{c_char, CString};
+use std::{
+    ffi::{c_char, CString},
+    process::{ExitCode, Termination},
+};
 
-use commands::cmd_completions::cmd_completions;
-use commands::cmd_list::cmd_list;
-use commands::cmd_mount::cmd_mount;
-use commands::cmd_subvolume::cmd_subvolumes;
-use commands::logger::SimpleLogger;
 use bch_bindgen::c;
+use log::debug;
 
 #[derive(Debug)]
 pub struct ErrnoError(pub errno::Errno);
@@ -29,13 +29,10 @@ fn handle_c_command(mut argv: Vec<String>, symlink_cmd: Option<&str>) -> i32 {
 
     let argc: i32 = argv.len().try_into().unwrap();
 
-    let argv: Vec<_> = argv
-        .into_iter()
-        .map(|s| CString::new(s).unwrap())
-        .collect();
+    let argv: Vec<_> = argv.into_iter().map(|s| CString::new(s).unwrap()).collect();
     let mut argv = argv
         .into_iter()
-        .map(|s| Box::into_raw(s.into_boxed_c_str()) as *mut c_char)
+        .map(|s| Box::into_raw(s.into_boxed_c_str()).cast::<c_char>())
         .collect::<Box<[*mut c_char]>>();
     let argv = argv.as_mut_ptr();
 
@@ -45,7 +42,7 @@ fn handle_c_command(mut argv: Vec<String>, symlink_cmd: Option<&str>) -> i32 {
             "--help" => {
                 c::bcachefs_usage();
                 0
-            },
+            }
             "data" => c::data_cmds(argc, argv),
             "device" => c::device_cmds(argc, argv),
             "dump" => c::cmd_dump(argc, argv),
@@ -59,18 +56,18 @@ fn handle_c_command(mut argv: Vec<String>, symlink_cmd: Option<&str>) -> i32 {
             "mkfs" => c::cmd_format(argc, argv),
             "remove-passphrase" => c::cmd_remove_passphrase(argc, argv),
             "reset-counters" => c::cmd_reset_counters(argc, argv),
-            "set-option" => c::cmd_set_option(argc, argv),
+            "set-fs-option" => c::cmd_set_option(argc, argv),
             "set-passphrase" => c::cmd_set_passphrase(argc, argv),
-            "setattr" => c::cmd_setattr(argc, argv),
+            "set-file-option" => c::cmd_setattr(argc, argv),
             "show-super" => c::cmd_show_super(argc, argv),
             "unlock" => c::cmd_unlock(argc, argv),
             "version" => c::cmd_version(argc, argv),
 
-            #[cfg(fuse)]
+            #[cfg(feature = "fuse")]
             "fusemount" => c::cmd_fusemount(argc, argv),
 
             _ => {
-                println!("Unknown command {}", cmd);
+                println!("Unknown command {cmd}");
                 c::bcachefs_usage();
                 1
             }
@@ -78,7 +75,7 @@ fn handle_c_command(mut argv: Vec<String>, symlink_cmd: Option<&str>) -> i32 {
     }
 }
 
-fn main() {
+fn main() -> ExitCode {
     let args: Vec<String> = std::env::args().collect();
 
     let symlink_cmd: Option<&str> = if args[0].contains("mkfs") {
@@ -96,28 +93,29 @@ fn main() {
     if symlink_cmd.is_none() && args.len() < 2 {
         println!("missing command");
         unsafe { c::bcachefs_usage() };
-        std::process::exit(1);
+        return ExitCode::from(1);
     }
 
     unsafe { c::raid_init() };
 
-    log::set_boxed_logger(Box::new(SimpleLogger)).unwrap();
-    log::set_max_level(log::LevelFilter::Warn);
-
     let cmd = match symlink_cmd {
         Some(s) => s,
         None => args[1].as_str(),
     };
 
-    let ret = match cmd {
-        "completions" => cmd_completions(args[1..].to_vec()),
-        "list" => cmd_list(args[1..].to_vec()),
-        "mount" => cmd_mount(args, symlink_cmd),
-        "subvolume" => cmd_subvolumes(args[1..].to_vec()),
-        _ => handle_c_command(args, symlink_cmd),
-    };
+    match cmd {
+        "completions" => {
+            commands::completions(args[1..].to_vec());
+            ExitCode::SUCCESS
+        }
+        "list" => commands::list(args[1..].to_vec()).report(),
+        "mount" => commands::mount(args, symlink_cmd),
+        "subvolume" => commands::subvolume(args[1..].to_vec()).report(),
+        _ => {
+            let r = handle_c_command(args, symlink_cmd);
 
-    if ret != 0 {
-        std::process::exit(1);
+            debug!("return code from C command: {r}");
+            ExitCode::from(r as u8)
+        }
     }
 }
